@@ -43,10 +43,9 @@
 //! ```
 //! use reql::r;
 //!
-//! # async fn connect() -> reql::Result<()> {
-//! let connection = r.connect(()).await?;
-//! # Ok(())
-//! # }
+//! # async fn example() -> reql::Result<()> {
+//! let session = r.connect(()).await?;
+//! # Ok(()) };
 //! ```
 //!
 //! The variable `connection` is now initialized and we can run queries.
@@ -54,15 +53,9 @@
 //! # Send a query to the database #
 //!
 //! ```
-//! use futures::TryStreamExt;
-//! use reql::r;
-//!
-//! # async fn connect() -> reql::Result<()> {
-//! let conn = r.connect(()).await?;
-//! let mut query = r.expr("Hello world!").run(&conn);
-//! # let _: Option<String> = query.try_next().await?;
-//! # Ok(())
-//! # }
+//! # reql::example(|r, conn| async_stream::stream! {
+//! r.expr("Hello world!").run(conn)
+//! # });
 //! ```
 //!
 //! [See the `r` struct for more available commands](r)
@@ -75,39 +68,147 @@ mod proto;
 
 use async_net::TcpStream;
 use cmd::run::Response;
+use cmd::StaticString;
 use dashmap::DashMap;
-use futures::channel::mpsc::UnboundedSender;
+use futures::channel::mpsc::{self, UnboundedReceiver, UnboundedSender};
 use futures::lock::Mutex;
+use log::trace;
+use proto::Payload;
+use ql2::query::QueryType;
 use ql2::response::ResponseType;
 use ql2::term::TermType;
+use serde_json::json;
 use std::borrow::Cow;
+use std::ops::Drop;
 use std::sync::atomic::{AtomicBool, AtomicU64, Ordering};
+use std::sync::Arc;
+use types::ServerInfo;
 
 pub use err::*;
 pub use proto::Query;
+#[doc(inline)]
+pub use reql_types as types;
 
 /// Custom result returned by various ReQL commands
 pub type Result<T> = std::result::Result<T, Error>;
 
+type Sender = UnboundedSender<Result<(ResponseType, Response)>>;
+type Receiver = UnboundedReceiver<Result<(ResponseType, Response)>>;
+
 /// The connection object returned by `r.connect()`
 #[derive(Debug)]
-pub struct Connection {
+pub struct Session {
     db: Cow<'static, str>,
     stream: Mutex<TcpStream>,
-    channels: DashMap<u64, UnboundedSender<Result<(ResponseType, Response)>>>,
+    channels: DashMap<u64, Sender>,
     token: AtomicU64,
     broken: AtomicBool,
     change_feed: AtomicBool,
 }
 
-impl Connection {
+impl Session {
+    fn token(&self) -> u64 {
+        let token = self
+            .token
+            .fetch_update(Ordering::SeqCst, Ordering::SeqCst, |x| Some(x + 1))
+            .unwrap();
+        if token == u64::MAX {
+            self.mark_broken();
+        }
+        token
+    }
+
+    pub fn connection(&self) -> Result<Connection<'_>> {
+        self.broken()?;
+        self.change_feed()?;
+        let token = self.token();
+        let (tx, rx) = mpsc::unbounded();
+        self.channels.insert(token, tx);
+        Ok(Connection::new(self, rx, token))
+    }
+
+    /// Change the default database on this connection
+    ///
+    /// ## Example
+    ///
+    /// Change the default database so that we don’t need to specify the
+    /// database when referencing a table.
+    ///
+    /// ```
+    /// # reql::example(|r, conn| async_stream::stream! {
+    /// conn.r#use("marvel");
+    /// r.table("heroes").run(conn) // refers to r.db("marvel").table("heroes")
+    /// # });
+    /// ```
+    ///
+    /// ## Related commands
+    /// * [connect](r::connect)
+    /// * [close](Connection::close)
+    pub fn r#use<T>(&mut self, db_name: T)
+    where
+        T: StaticString,
+    {
+        self.db = db_name.static_string();
+    }
+
+    /// Ensures that previous queries with the `noreply` flag have been
+    /// processed by the server
+    ///
+    /// Note that this guarantee only applies to queries run on the given
+    /// connection.
+    ///
+    /// ## Example
+    ///
+    /// We have previously run queries with [noreply](cmd::run::Options::noreply())
+    /// set to `true`. Now wait until the server has processed them.
+    ///
+    /// ```
+    /// # async fn example() -> reql::Result<()> {
+    /// # let session = reql::r.connect(()).await?;
+    /// session.noreply_wait().await
+    /// # }
+    /// ```
+    ///
+    pub async fn noreply_wait(&self) -> Result<()> {
+        let mut conn = self.connection()?;
+        let payload = Payload(QueryType::NoreplyWait, None, Default::default());
+        trace!(
+            "waiting for noreply operations to finish; token: {}",
+            conn.token
+        );
+        let (typ, _) = conn.request(&payload, false).await?;
+        trace!(
+            "session.noreply_wait() run; token: {}, response type: {:?}",
+            conn.token,
+            typ,
+        );
+        Ok(())
+    }
+
+    pub async fn server(&self) -> Result<ServerInfo> {
+        let mut conn = self.connection()?;
+        let payload = Payload(QueryType::ServerInfo, None, Default::default());
+        trace!("retrieving server information; token: {}", conn.token);
+        let (typ, resp) = conn.request(&payload, false).await?;
+        trace!(
+            "session.server() run; token: {}, response type: {:?}",
+            conn.token,
+            typ,
+        );
+        let mut vec = serde_json::from_value::<Vec<ServerInfo>>(resp.r)?;
+        let info = vec
+            .pop()
+            .ok_or_else(|| Driver::Other("server info is empty".into()))?;
+        Ok(info)
+    }
+
     fn mark_broken(&self) {
         self.broken.store(true, Ordering::SeqCst);
     }
 
     fn broken(&self) -> Result<()> {
         if self.broken.load(Ordering::SeqCst) {
-            return Err(err::Client::ConnectionBroken.into());
+            return Err(err::Driver::ConnectionBroken.into());
         }
         Ok(())
     }
@@ -125,14 +226,87 @@ impl Connection {
         self.change_feed.store(false, Ordering::SeqCst);
     }
 
+    fn is_change_feed(&self) -> bool {
+        self.change_feed.load(Ordering::SeqCst)
+    }
+
     fn change_feed(&self) -> Result<()> {
         if self.change_feed.load(Ordering::SeqCst) {
-            return Err(err::Client::ConnectionLocked.into());
+            return Err(err::Driver::ConnectionLocked.into());
         }
         Ok(())
     }
 }
 
+#[derive(Debug, Clone)]
+pub struct Connection<'a> {
+    session: &'a Session,
+    rx: Arc<Mutex<Receiver>>,
+    token: u64,
+}
+
+impl<'a> Connection<'a> {
+    fn new(session: &'a Session, rx: Receiver, token: u64) -> Connection<'a> {
+        Connection {
+            session,
+            token,
+            rx: Arc::new(Mutex::new(rx)),
+        }
+    }
+
+    /// Close an open connection
+    ///
+    /// ## Example
+    ///
+    /// Close an open connection, waiting for noreply writes to finish.
+    ///
+    /// ```
+    /// # async fn example() -> reql::Result<()> {
+    /// # let session = reql::r.connect(()).await?;
+    /// # let mut conn = session.connection()?;
+    /// conn.close(()).await
+    /// # }
+    /// ```
+    ///
+    /// [Read more about this command →](cmd::close)
+    pub async fn close<T>(&mut self, arg: T) -> Result<()>
+    where
+        T: cmd::close::Arg,
+    {
+        if !self.session.is_change_feed() {
+            trace!(
+                "ignoring conn.close() called on a normal connection; token: {}",
+                self.token
+            );
+            return Ok(());
+        }
+        let arg = if arg.noreply_wait() {
+            None
+        } else {
+            Some(r.expr(json!({ "noreply": false })))
+        };
+        let payload = Payload(QueryType::Stop, arg, Default::default());
+        trace!("closing a changefeed; token: {}", self.token);
+        let (typ, _) = self.request(&payload, false).await?;
+        self.session.unmark_change_feed();
+        trace!(
+            "conn.close() run; token: {}, response type: {:?}",
+            self.token,
+            typ,
+        );
+        Ok(())
+    }
+}
+
+impl Drop for Connection<'_> {
+    fn drop(&mut self) {
+        self.session.channels.remove(&self.token);
+        if self.session.is_change_feed() {
+            self.session.unmark_change_feed();
+        }
+    }
+}
+
 /// The top-level ReQL namespace
 ///
 /// # Example
@@ -153,17 +327,15 @@ impl r {
     /// Open a connection using the default host and port, specifying the default database.
     ///
     /// ```
-    /// use reql::cmd::connect::Options;
-    /// use reql::r;
+    /// use reql::{r, cmd::connect::Options};
     ///
-    /// # async fn connect() -> reql::Result<()> {
-    /// let conn = r.connect(Options::new().db("marvel")).await?;
-    /// # Ok(())
-    /// # }
+    /// # async fn example() -> reql::Result<()> {
+    /// let session = r.connect(Options::new().db("marvel")).await?;
+    /// # Ok(()) }
     /// ```
     ///
-    /// Read more about this command [cmd::connect]
-    pub async fn connect<T>(self, options: T) -> Result<Connection>
+    /// Read more about this command [connect](cmd::connect)
+    pub async fn connect<T>(self, options: T) -> Result<Session>
     where
         T: cmd::connect::Arg,
     {
@@ -192,21 +364,16 @@ impl r {
     ///
     /// The `db` command is optional. If it is not present in a query, the
     /// query will run against the default database for the connection,
-    /// specified in the `db` argument to [r::connect].
+    /// specified in the `db` argument to [connect](r::connect).
     ///
     /// # Examples
     ///
     /// Explicitly specify a database for a query.
     ///
     /// ```
-    /// # use futures::TryStreamExt;
-    /// # use reql::r;
-    /// # async fn example() -> reql::Result<()> {
-    /// # let conn = r.connect(()).await?;
-    /// let mut query = r.db("heroes").table("marvel").run(&conn);
-    /// # let _: Option<String> = query.try_next().await?;
-    /// # Ok(())
-    /// # }
+    /// # reql::example(|r, conn| async_stream::stream! {
+    /// r.db("heroes").table("marvel").run(conn)
+    /// # });
     /// ```
     pub fn db<T>(self, arg: T) -> Query
     where
@@ -507,3 +674,13 @@ impl r {
         arg.into_query()
     }
 }
+
+// Helper for making writing examples less verbose
+#[doc(hidden)]
+pub fn example<'a, Q, F, S>(_query: Q)
+where
+    Q: FnOnce(r, &'a mut Session) -> async_stream::AsyncStream<(), F>,
+    F: futures::Future<Output = S>,
+    S: futures::Stream<Item = Result<serde_json::Value>>,
+{
+}
